package de.hdm_stuttgart.mi.sd1.test.ignore_me;

import org.junit.Assert;

import java.lang.reflect.*;
import java.util.Arrays;
import java.util.stream.Collectors;

public class ObjectWrapper<T> {

  private final Class classz;
  private final T instance;

  static public void assertFinal(final Class classZ, final String fieldName) {
    try {
      final int modifier = classZ.getDeclaredField(fieldName).getModifiers();
      Assert.assertTrue("»" + fieldName + "« can be modified, no immutable!", Modifier.isFinal(modifier));

    } catch (NoSuchFieldException e) {
      Assert.fail("No such field »" + fieldName + "«");
    }
  }

  public ObjectWrapper(final Class<T> classz, final Object ... parameter) {
    this.classz = classz;
    T tmpInstance = null;
    final Constructor<?>[] candidates = Arrays.
            stream(classz.getConstructors()).
            filter(c -> matchesArgumentList(c, parameter)).
            toArray(Constructor<?>[]::new);
    if (0 == candidates.length) {
      Assert.fail("No suitable constructor in class »" + classz.getName() + "« matching arguments »(" +
              Arrays.stream(parameter).map(Object::toString).collect(Collectors.joining(", ")) +
              ")«");
    } else if (1 < candidates.length) {
      Assert.fail("Multiple constructor matches due to ambiguous signature");
    } else {
      try {
        tmpInstance = (T) candidates[0].newInstance(
                Arrays.stream(parameter).toArray());
      } catch (Exception e) {
        Assert.fail("Unable to instantiate instance of class »" + classz.getName() + "«: "+ e);
      }
    }
    instance = tmpInstance;
  }

  public <R> void assertFieldExists(final Class<R> valueType, final String name) {
    try {
      final Field field = classz.getField(name);
      if (!valueType.equals(field.getType())) {
        Assert.fail("Field »" + name + "« in class »" + classz.getName() +  "« is not of type »" +
                valueType.getName() + "«");
      }
    } catch (final NoSuchFieldException e) {
      Assert.fail("No such field »" + name + "« in class »" + classz.getName() + "«");
    }
  }

  public <R> R get(final Class<R> valueType, final String name) {

    try {
      final Field field = classz.getField(name);
      if (valueType.equals(field.getType())) {
        return (R) field.get(instance);
      } else {
        Assert.fail("Field »" + name + "« in class »" + classz.getName() +  "« is not of type »" +
                valueType.getName() + "«");
      }
    } catch (final NoSuchFieldException e) {
      Assert.fail("No such field »" + name + "« in class »" + classz.getName() + "«");
    } catch (final IllegalAccessException e) {
      Assert.fail("Unable to access field »" + name + "« in class »" + classz.getName() + "«: " + e.getMessage());
    }
    return null;
  }
  public void set(final String name, final Object value) {

    try {
      final Field field = classz.getField(name);
      final Class  argumentType = value.getClass();
      if (field.getType().equals(argumentType) ||
              field.getType().equals(getPrimitiveType(argumentType))){
       field.set(instance, value);
      } else {
        Assert.fail("Field »" + name + "« in class »" + classz.getName() +  "« is not of type »" +
                argumentType.getName() + "«");
      }
    } catch (final NoSuchFieldException e) {
      Assert.fail("No such field »" + name + "« in class »" + classz.getName() + "«");
    } catch (final IllegalAccessException e) {
      Assert.fail("Unable to access field »" + name + "« in class »" + classz.getName() + "«: " + e.getMessage());
    }
  }

  public <R> R invoke(final Class<R> returnType, final String name, final Object ... parameter) {

    final Method[] candidates = Arrays.
            stream(classz.getMethods()).
            filter(m-> m.getName().equals(name) && matchesArgumentList(m, parameter)).
            toArray(Method[]::new);

    if (0 == candidates.length) {
      Assert.fail("No suitable method found");
    } else if (1 < candidates.length) {
      Assert.fail("Multiple method matches due to ambiguous signature");
    } else {
      final Method method = candidates[0];
      if (method.getReturnType().equals(returnType)) {
        try {
          return (R) method.invoke(instance, parameter);
        } catch (final Exception e) {
          Assert.fail("Unable to execute method: " + e);
        }
      } else {
        Assert.fail("Method »" + method.getName() + "« does have return type »"  + method.getReturnType() + "«" +
                "rather then »" + returnType.getName() + "«");
      }
    }
    return null;
  }

  /**
   * Check for a given array of objects matching an {@link Executable}'s argument list.
   *
   * @param executable
   * @param parameter
   * @return <code>true</code> if parameters match the {@link Executable}'s argument list, <code>false</code>
   * otherwise.
   */
  private boolean matchesArgumentList(final Executable executable, final Object ... parameter) {

    if (executable.getParameterCount() != parameter.length) {
      return false;
    } else {
      final Class<?>[] formalArgumentTypes = executable.getParameterTypes();
      for (int i = 0; i < formalArgumentTypes.length; i++) {
        final Class parametersClass = parameter[i].getClass();

        final Class<?> formalArgumentType = formalArgumentTypes[i];

        if (formalArgumentType.isPrimitive()) {
          if (!formalArgumentType.equals(getPrimitiveType(parametersClass))) {
            return false;
          }
        } else {
          if (!formalArgumentType.isAssignableFrom(parametersClass)) {
            return false;
          }
        }
      }
    }
    return true;
  }

  static private Class<?> getPrimitiveType (final Class<?> in) {
    if (in.equals(Byte.class)) {
      return byte.class;
    } else if (in.equals(Short.class)) {
      return short.class;
    } else if (in.equals(Integer.class)) {
      return int.class;
    } else if (in.equals(Long.class)) {
      return long.class;
    } else if (in.equals(Float.class)) {
      return float.class;
    } else if (in.equals(Double.class)) {
      return double.class;
    } else if (in.equals(Boolean.class)) {
      return boolean.class;
    } else if (in.equals(Character.class)) {
      return char.class;
    } else {
      return in;            // Type is no primitive
    }
  }
}
